import os
from .. project import aliases, construct, load, project
from . import animation, failed, runner
from .. util import log


class Collection(animation.Animation):
    """
    A ``Collection`` is a list of ``Animation``s
    """

    @staticmethod
    def pre_recursion(desc):
        animations = []
        name_count = {}

        for a in desc['animations']:
            loaded = load.load_if_filename(a)
            if loaded:
                animation = loaded.get('animation', {})
                run = loaded.get('run', {})

            elif callable(a) or isinstance(a, str) or 'animation' not in a:
                animation = a
                run = {}

            elif not isinstance(a, dict):
                raise TypeError('Unexpected type %s in collection' % type(a))

            elif 'typename' in a:
                run = a.pop('run', {})
                animation = a

            else:
                animation = a.pop('animation', {})
                run = a.pop('run', {})
                if a:
                    raise ValueError('Extra animation fields: ' + ', '.join(a))

            animation = construct.to_type_constructor(
                animation, 'bibliopixel.animation')

            animation['run'] = run

            datatype = animation.setdefault('datatype', failed.Failed)
            name = animation.setdefault('name', datatype.__name__)
            name_count[name] = 1 + name_count.get(name, 0)

            # Children without fps or sleep_time get it from their parents.
            if not ('fps' in run or 'sleep_time' in run):
                if 'fps' in desc['run']:
                    run.update(fps=desc['run']['fps'])
                elif 'sleep_time' in desc['run']:
                    run.update(sleep_time=desc['run']['sleep_time'])

            animations.append(animation)

        dupes = {k: 1 for k, v in name_count.items() if v > 1}
        for a in animations:
            name = a['name']
            count = dupes.get(name)
            if count:
                dupes[name] += 1
                a['name'] += '_' + str(count - 1)

        desc['animations'] = animations
        return desc

    CHILDREN = 'animations',

    def __init__(self, layout, animations=None, **kwds):
        super().__init__(layout, **kwds)
        self.animations = _AnimationList(animations or [])
        self.internal_delay = 0  # never wait

    # Override to handle all the animations
    def cleanup(self, clean_layout=True):
        self.state = runner.STATE.canceled
        for a in self.animations:
            a.cleanup()
        super().cleanup(clean_layout)

    def add_animation(self, anim, **kwds):
        from .. util import deprecated
        deprecated.deprecated('Collection.add_animation')

        anim._set_runner(kwds)
        self.animations.append(anim)

    def pre_run(self):
        for a in self.animations:
            a.pre_run()

    def set_project(self, project):
        super().set_project(project)
        for a in self.animations:
            a.set_project(project)

    def detach(self, overlay):
        """
        Give each animation a unique, mutable layout so they can run
        independently.
        """
        # See #868
        #
        for i, a in enumerate(self.animations):
            a.layout = a.layout.clone()
            if overlay and i:
                a.preclear = False


class _AnimationList:
    def __init__(self, animations):
        self._animations = list(animations)
        self._names = {a.name: i for i, a in enumerate(self._animations)}

    def _index(self, i):
        if isinstance(i, int):
            return i % len(self._animations)
        return self._names[i]

    def append(self, animation):
        # We're forced to do the "fill in a name and make sure it's
        # unique" in two different ways, because of this method.
        #
        # TODO: We should disallow APIs from doing surgery on
        # self.animations after construction.  Right now only
        # Remote does this, or legacy code that calls add_animation.
        # That code should go on ``cls.pre_recursion()``.
        self._animations.append(animation)
        try:
            base_name = animation.name
        except AttributeError:
            base_name = animation.__class__.__name__

        count = 0
        name = base_name

        while name in self._names:
            name = '%s_%d' % (base_name, count)
            count += 1

        animation.name = name

    def __getitem__(self, i):
        return self._animations[self._index(i)]

    def __getattr__(self, i):
        return self[i]

    def __iter__(self):
        return iter(self._animations)

    def __len__(self):
        return len(self._animations)



from ... util import pointOnCircle
import math


def calc_ring_pixel_count(rings):
    num = 0
    for r in rings:
        num += len(r)
    return num


def calc_ring_steps(rings):
    steps = []
    for r in rings:
        steps.append(360.0 / len(r))
    return steps


def make_circle_coord_map(rings=None, pixels_per=None, offset=0, invert=False):
    if pixels_per:
        rings = []
        for c in pixels_per:
            rings.append([offset, offset + c])
            offset = offset + c

    if not rings:
        raise ValueError('Must specify rings or pixels_per')

    if rings:
        num = 0
        out_rings = []
        for r in rings:
            if len(r) != 2:
                raise ValueError(
                    '"rings" values must only be first and last index.')
            if r[0] < r[1]:
                indices = list(range(r[0], r[1]))
            else:
                indices = list(range(r[1], r[0]))[::-1]
            out_rings.append(indices)
            num += len(indices)

        if invert:
            out_rings = out_rings[::-1]

        return (out_rings, calc_ring_steps(out_rings))


def make_circle_coord_map_positions(
        rings, origin=(0, 0, 0), z_diff=0, radii=None):
    if len(origin) not in [2, 3]:
        raise ValueError('origin must be (x,y) or (x,y,z)')

    use_z = len(origin) == 3
    if use_z:
        ox, oy, oz = origin
    else:
        ox, oz = origin

    num = calc_ring_pixel_count(rings)
    steps = calc_ring_steps(rings)

    points = [None] * num
    z = 0

    for i in range(len(rings)):
        r = rings[i]
        step = steps[i]
        angle = 0.0
        for p in r:
            if radii:
                radius = radii[i]
            else:
                radius = (len(r) - 1) * math.pi
            x, y = pointOnCircle(0, 0, radius, angle)
            if use_z:
                points[p] = (x + ox, y + oy, z + oz)
            else:
                points[p] = (x + ox, y + oy)

            angle += step

        z += z_diff

    return points



from . animation import Animation


class Off(Animation):
    """A trivial animation that turns all pixels in a layout off."""

    def __init__(self, layout, timeout=1, **kwds):
        super().__init__(layout, **kwds)
        self.internal_delay = timeout

    def step(self, amt=1):
        self.layout.all_off()


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    OffAnim = Off



from . import make


def pop_legacy_palette(kwds, *color_defaults):
    """
    Older animations in BPA and other areas use all sorts of different names for
    what we are now representing with palettes.

    This function mutates a kwds dictionary to remove these legacy fields and
    extract a palette from it, which it returns.
 """
    palette = kwds.pop('palette', None)
    if palette:
        legacy = [k for k, _ in color_defaults if k in kwds]
        if legacy:
            raise ValueError('Cannot set palette and ' + ', '.join(legacy))
        return palette

    values = [kwds.pop(k, v) for k, v in color_defaults]
    if values and color_defaults[0][0] in ('colors', 'palette'):
        values = values[0]

    return make.colors(values or None)



from .. project import construct
from . driver_base import DriverBase

import copy


class Mirror(DriverBase):
    @staticmethod
    def pre_recursion(desc):
        if 'drivers' not in desc:
            raise ValueError('MirrorDriver must have a drivers section')

        DriverBase.pre_recursion(desc)

        old_desc, desc = desc, {}
        for name in 'typename', 'datatype', 'drivers':
            value = old_desc.pop(name, None)
            if value is not None:
                desc[name] = value

        desc['drivers'] = [construct.to_type(d) for d in desc['drivers']]
        desc['drivers'] = [dict(d, **old_desc) for d in desc['drivers']]
        desc['num'] = old_desc['num']
        return desc

    CHILDREN = 'drivers',

    def __init__(self, num=0, *, drivers, **kwds):
        super().__init__(num, **kwds)
        self.drivers = drivers

        def forward(name):
            def method(*args, **kwds):
                for d in self.drivers:
                    getattr(d, name)(*args, **kwds)

            return method

        self.cleanup = forward('cleanup')
        self.set_colors = forward('set_colors')
        self.set_pixel_positions = forward('set_pixel_positions')
        self.start = forward('start')
        self.sync = forward('sync')
        self.update_colors = forward('update_colors')


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    MirrorDriver = Mirror



"""
This is probably not used and should likely be replaced by the code
in BiblioPixelAnimations.matrix.ImageAnim
"""
from .. import deprecated


def convert_mode(image, mode='RGB'):
    """Return an image in the given mode."""
    deprecated.deprecated('util.gif.convert_model')

    return image if (image.mode == mode) else image.convert(mode=mode)


def image_to_colorlist(image, container=list):
    """Given a PIL.Image, returns a ColorList of its pixels."""
    deprecated.deprecated('util.gif.image_to_colorlist')

    return container(convert_mode(image).getdata())


def animated_gif_to_colorlists(image, container=list):
    """
    Given an animated GIF, return a list with a colorlist for each frame.
    """
    deprecated.deprecated('util.gif.animated_gif_to_colorlists')

    from PIL import ImageSequence

    it = ImageSequence.Iterator(image)
    return [image_to_colorlist(i, container) for i in it]



"""
Simple numerical operations put together in a list, useful for
scaling or offsetting some incoming numerical value.

Examples:

   Ops()(x)

returns the number unchanged

   Ops('mul', 0.5)(x)

multiples x by 0.5

   Ops('add', 0.1)(x)

adds 1 to x

   Ops('sqrt')(x)

returns the square root of x

   Ops('mul', 0.5, 'add', 1, 'sqrt')(x)

returns sqrt(1 + 0.5 * x)

"""

import math, operator


class Op:
    def __init__(self, op, *values):
        if len(values) > len(OPS):
            raise KeyError('Too many arguments for ' + op)
        ops = OPS[len(values)]

        if op not in ops:
            raise KeyError(OPS_ERROR.format(op, ', '.join(sorted(ops))))
        self.op = ops[op]

        self.opname = op
        self.values = values

    def __call__(self, x):
        return self.op(x, *self.values)

    def __str__(self):
        values = ', '.join(str(v) for v in self.values)
        return '%s(%s)' % self.opname, values


class Ops:
    def __init__(self, *ops):
        self.ops = tuple(_make_ops(ops))

    def __call__(self, x):
        for op in self.ops:
            x = op(x)
        return x

    def __bool__(self):
        return bool(self.ops)

    def __str__(self):
        return '->'.join(str(i) for i in self.ops)


def _make_ops(ops):
    op, values = None, []
    for o in ops:
        if isinstance(o, str):
            # We're starting a new operation
            if op:
                yield Op(op, *values)
            op, values = o, []
        else:
            # We're in the middle of an operation
            if not op:
                raise ValueError('Value without op')
            values.append(o)
    if op:
        yield Op(op, *values)


def _two_ops():
    def add(ops, op):
        f = getattr(operator, 'truediv' if op == 'div' else op)
        ops[op] = f
        if op not in COMMUTATIVE:
            ops['r' + op] = lambda x, y: f(y, x)

    ops = {}
    for op in TWO_OPERATORS:
        add(ops, op)

    ops.update(max=max, min=min, gamma=ops['rpow'])
    return ops


def _join(items, indent='    ', per_line=8):
    lines = []
    for i in range(0, len(items), 8):
        lines.append(indent + ', '.join(items[i:i + 8]) + '\n')
    return ''.join(lines)


TWO_OPERATORS = 'add', 'div', 'mod', 'mul', 'pow', 'sub'
COMMUTATIVE = 'add', 'mul'

ONE_OPERATORS = (
    'acos',
    'acosh',
    'asin',
    'asinh',
    'atan',
    'atanh',
    'ceil',
    'cos',
    'cosh',
    'degrees',
    'erf',
    'erfc',
    'exp',
    'expm1',
    'fabs',
    'floor',
    'gamma',
    'lgamma',
    'log',
    'log10',
    'log1p',
    'log2',
    'radians',
    'sin',
    'sinh',
    'sqrt',
    'tan',
    'tanh',
    'trunc')

OPS = [{k: getattr(math, k) for k in ONE_OPERATORS}, _two_ops()]

OPS_ERROR = """\
Do not understand operator {0}.

One argument operators are:
   %s

Two argument operators are:
   %s
""" % tuple(_join(sorted(o)) for o in OPS)



import os
from .. import data_file, log

CODE_BLOCK = '.. code-block:: '

IMAGE = '.. image:: \
https://raw.githubusercontent.com/ManiacalLabs/DocsFiles/master/BiblioPixel/'

LANGS = 'json', 'yaml'


def extract_gif_lines(lines):
    for filename, code in _extract(lines):
        codelines = '\n'.join(code)
        try:
            project = data_file.loads(codelines)
        except:
            log.error('Unable to load code: $s\n%s', filename, codelines)
        else:
            yield filename, project


def _extract(lines):
    in_code = False

    for line in lines:
        if line.startswith('.. '):
            if line.startswith(CODE_BLOCK):
                code = []
                in_code = True

            elif line.startswith(IMAGE) and in_code:
                filename = line[len(IMAGE):].strip()
                code = _remove_common_prefix(code)
                yield filename, code
                in_code = False

            elif line.startswith(IMAGE) and not in_code:
                pass

            elif line.startswith('.. image:'):
                log.printer('missing image')
                log.printer(line)
                log.printer(IMAGE)

        elif in_code:
            if not line or line[0].isspace():
                code.append(line)
            else:
                in_code = False


def _remove_common_prefix(lines):
    lines = list(lines)
    while lines and not lines[0].strip():
        lines.pop(0)

    while lines and not lines[-1].strip():
        lines.pop()

    fill = ' ' * max(len(i) for i in lines)
    lines = [i or fill for i in lines]

    prefix = os.path.commonprefix(lines)
    spaces = len(prefix) - len(prefix.lstrip())
    return [i[spaces:] for i in lines]



import fractions
from . import control
from .. util import log

try:
    import mido
    MESSAGE_TYPES = set(s['type'] for s in mido.messages.specs.SPECS)

except:
    mido = None
    MESSAGE_TYPES = set()

MIDO_ERROR = """You need to install mido.  Try
    pip install mido
"""

MIDO_BACKEND_ERROR = """You have the wrong rtmidi installed.  Try
    pip uninstall -y rtmidi
    pip install -y python-rtmidi
"""


class Midi(control.ExtractedLoop):
    EXTRACTOR = {
        # There are more mido message types that we haven't used yet.
        'keys_by_type': {
            'aftertouch': (
                'port', 'channel', 'type', 'value'),
            'control_change': (
                'port', 'channel', 'type', 'control', 'value'),
            'note_off': (
                'port', 'channel', 'type', 'note', 'velocity'),
            'note_on': (
                'port', 'channel', 'type', 'note', 'velocity'),
            'pitchwheel': (
                'port', 'channel', 'type', 'pitch'),
            'program_change': (
                'port', 'channel', 'type', 'program'),
        },

        # Some numeric fields (channel, control, program and note) don't get
        # normalized because they are basically names.
        'normalizers': {
            'pitch': lambda x: fractions.Fraction(x - 8192) / 8192,
            'value': lambda x: fractions.Fraction(x) / 127,
            'velocity': lambda x: fractions.Fraction(x) / 127,
        },

        'omit': ('port', 'channel'),
    }

    def __init__(self, use_note_off=True, **kwds):
        """
        :param use_note_off:
            If False, map note_offs to note_ons with velocity 0
            If True, map note_ons with velocity 0 to note_offs
            If None, do not change none_ons or note_offs
        """
        super().__init__(**kwds)
        self.use_note_off = use_note_off

    def messages(self):
        if not mido:
            raise ValueError(MIDO_ERROR)
        try:
            input_names = mido.get_input_names()
        except AttributeError as e:
            e.args = (MIDO_ERROR,) + e.args
            raise

        ports = [mido.open_input(i) for i in input_names]

        if not ports:
            log.error('control.midi: no MIDI ports found')
            return

        port_names = ', '.join('"%s"' % p.name for p in ports)
        log.info('Starting to listen for MIDI on port%s %s',
                 '' if len(ports) == 1 else 's', port_names)
        for port, msg in mido.ports.MultiPort(ports, yield_ports=True):
            mdict = dict(vars(msg), port=port)

            if self.use_note_off:
                if msg.type == 'note_on' and not msg.velocity:
                    mdict.update(type='note_off')
            elif self.use_note_off is False:
                if msg.type == 'note_off':
                    mdict.update(type='note_on', velocity=0)

            yield mdict



import copy, ctypes, enum
from . server_driver import ServerDriver
from .. util import artnet_message, log, offset_range, server_cache, udp


class ArtNet(ServerDriver):
    SERVER_CLASS = udp.Sender

    def __init__(self, *args, ip_address='', port=artnet_message.UDP_PORT,
                 filter_dupes=True, offset=0, **kwds):
        """
        :param dict channel_map: maps DMX channels to positions in
            the color_list
        :param int offset: a DMX channel offset, positive, negative or zero
        """
        super().__init__(*args, address=(ip_address, port), **kwds)
        self.filter_dupes = filter_dupes
        self.offset = offset_range.DMXChannel(offset)
        self.msg = artnet_message.dmx_message()
        self.last_message = None

    def _send_packet(self):
        self._copy_buffer_into_data()
        msg = bytes(self.msg)
        if not (self.filter_dupes and msg == self.last_message):
            self.server.send(msg)
            self.last_message = msg

    def _copy_buffer_into_data(self):
        self.offset.copy_to(self._buf, self.msg.data)

    def _on_positions(self):
        pass



from . animation import Animation
from .. layout import Circle as CircleLayout
from .. util import deprecated


class Circle(Animation):
    LAYOUT_CLASS = CircleLayout
    LAYOUT_ARGS = 'rings',

    def __init__(self, layout, **kwds):
        super().__init__(layout, **kwds)
        self.rings = layout.rings
        self.ringCount = layout.ringCount
        if deprecated.allowed():  # pragma: no cover
            self.lastRing = layout.lastRing
        self.ringSteps = layout.ringSteps


if deprecated.allowed():  # pragma: no cover
    BaseCircleAnim = Circle



from .. project import construct
from .. util import deprecated, flatten
from . action import ActionList
from . receiver import Receiver


class Routing(Receiver):
    """
    A dict that routes a message to an ActionList.
    """

    def __init__(self, routing, default, python_path):
        """
        :param dict routing: `routing` is a dict that maps addresses
           to lists of actions.

           The values in the input dictionary `routing` are recursively visited
           to build the routing table:

           * values that are strings or lists are used to construct ActionLists
           * dictionaries that contain "typename" or "datatype" keys are
             used to construct a class of that type.
           * otherwise, dictionaries are visited recursively
           * all other types are forbidden
        """
        def make(x):
            if isinstance(x, (list, str)):
                return ActionList(x)

            assert isinstance(x, dict)
            if 'datatype' in x or 'typename' in x:
                x = dict(default, **x)
                return construct.construct_type(x, python_path)

            return {k: make(v) for k, v in x.items()}

        routing = flatten.unflatten(routing)
        self.routing = make(routing)

    def set_project(self, project):
        """Set the base project for routing."""
        def visit(x):
            # Try to set_project, then recurse through any values()
            set_project = getattr(x, 'set_project', None)
            if set_project:
                set_project(project)
            values = getattr(x, 'values', lambda: ())
            for v in values():
                visit(v)

        visit(self.routing)

    def receive(self, msg):
        """
        Returns a (receiver, msg) pair, where receiver is `None` if no route for
        the message was found, or otherwise an object with a `receive` method
        that can accept that `msg`.
        """
        x = self.routing
        while not isinstance(x, ActionList):
            if not x or not msg:
                return None, msg

            if not isinstance(x, dict):
                raise ValueError('Unexpected type %s' % type(x))

            _, value = msg.popitem(last=False)
            x = x.get(str(value))

        return x, msg

    def __bool__(self):
        return bool(self.routing)

    def __str__(self):
        return str(self.routing)

    if deprecated.allowed():
        set_root = set_project

        @property
        def root(self):
            return self.project



"""
Send an interrupt signal to a BiblioPixel process running on this
machine to kill it

DEPRECATED: use
.. code-block:: bash

    $ kill -int `bpa-pid`

"""

DESCRIPTION = """
Example:
`$ bp shutdown`

"""

from . import restart


def add_arguments(parser):
    restart.add_signal_arguments(parser, 'SIGKINT')



import collections


class Extractor:
    """
    Extractor is a class that extracts and normalizes values from
    incoming message dictionaries into ordered dictionaries based on the
    `type` key of each message.
    """

    def __init__(self, omit=None, normalizers=None, keys_by_type=None,
                 accept=None, reject=None, auto_omit=True):
        """
        Arguments

        omit -- A list of keys that will not be extracted.

        normalizers -- Some keys also need to be "normalized" -
            scaled and offset so they are between 0 and 1, or -1 and 1.
            The `normalizers` table maps key names to a function that
            normalizes the value of that key.

        keys_by_type -- `keys_by_type` is a dictionary from the `type` in an
            incoming message to a list of message keys to be extracted

        accept -- maps keys to a value or a list of values that are
            accepted for that key.  A message has to match *all* entries in
            `accept` to be accepted.

        reject -- map key to a value or a list of values that are not
            accepted for that key.  A message is rejected if it matches *any*
            entry in the reject map.

        auto_omit -- if True, omit all keys in `accept` that only have one
            possible value.

            auto_omit=True, the default, is probably more useful: if you
            request data for channel=1, type=note_on then you probably don't
            want to see channel=1, type=note_on with each message.
        """
        def to_set(x):
            if x is None:
                return set()
            if isinstance(x, (list, tuple)):
                return set(x)
            return set([x])

        def make_match(m):
            return m and {k: to_set(v) for k, v in m.items()}

        self.accept, self.reject = make_match(accept), make_match(reject)
        self.omit = to_set(omit)
        if auto_omit and self.accept:
            self.omit.update(k for k, v in self.accept.items() if len(v) == 1)

        self.normalizers = normalizers or {}
        if keys_by_type is None:
            self.keys_by_type = None
        else:
            self.keys_by_type = {}
            for k, v in keys_by_type.items():
                if isinstance(v, str):
                    v = [v]
                self.keys_by_type[k] = tuple(i for i in v if i not in self.omit)

    def extract(self, msg):
        """Yield an ordered dictionary if msg['type'] is in keys_by_type."""
        def normal(key):
            v = msg.get(key)
            if v is None:
                return v
            normalizer = self.normalizers.get(key, lambda x: x)
            return normalizer(v)

        def odict(keys):
            return collections.OrderedDict((k, normal(k)) for k in keys)

        def match(m):
            return (msg.get(k) in v for k, v in m.items()) if m else ()

        accept = all(match(self.accept))
        reject = any(match(self.reject))

        if reject or not accept:
            keys = ()
        elif self.keys_by_type is None:
            keys = [k for k in msg.keys() if k not in self.omit]
        else:
            keys = self.keys_by_type.get(msg.get('type'))
        return odict(keys)



import os, threading

import socketserver as SocketServer
from .. drivers.return_codes import RETURN_CODES
from . network import CMDTYPE
from .. util import log


class ThreadedDataHandler(SocketServer.BaseRequestHandler):

    def handle(self):
        try:
            cmd = ord(self.request.recv(1))
            r = bytes(self.request.recv(2))
            size = (ord(r[1]) << 8) | ord(r[0])

            if cmd == CMDTYPE.PIXEL_DATA:
                data = bytearray()
                empty_count = 0
                while len(data) < size:
                    buf = self.request.recv(4096)
                    if len(buf) == 0:
                        empty_count += 1
                    if empty_count >= 5:
                        log.exception(
                            "Failed to receive expected amount of data! "
                            "Expected: %s bytes / Received: %s bytes",
                            size, len(data))
                        self.request.sendall(RETURN_CODES.ERROR_SIZE)
                        return

                    data.extend(buf)

                if len(data) != size:
                    log.exception(
                        "Received data size incorrect! "
                        "Expected: %s bytes / Received: {%s bytes",
                        size, len(data))
                    return

                self.server.update(data)

                if self.server.hasFrame:
                    while self.server.hasFrame():
                        pass

                packet = bytearray()
                packet.append(RETURN_CODES.SUCCESS)
                self.request.sendall(packet)

            elif cmd == CMDTYPE.BRIGHTNESS:
                res = self.request.recv(1)
                bright = ord(res)
                result = RETURN_CODES.ERROR_UNSUPPORTED
                if self.server.set_brightness:
                    if self.server.set_brightness(bright):
                        result = RETURN_CODES.SUCCESS
                    else:
                        # Try again.
                        self.server.set_brightness(bright)

                packet = bytearray()
                packet.append(result)
                self.request.sendall(packet)

        except Exception as e:
            log.exception(e)
            pass  # if there's a comm error, just move on
        return


class ThreadedDataServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    update = None
    set_brightness = None
    hasFrame = None


class NetworkReceiver:

    def __init__(self, layout, port=3142, interface='0.0.0.0'):
        self.layout = layout
        self.address = (interface, port)
        SocketServer.TCPServer.allow_reuse_address = True
        self._server = ThreadedDataServer(self.address, ThreadedDataHandler)
        self._server.update = self.update
        self._server.set_brightness = self.layout.set_brightness

    def start(self, join=False):
        self._t = threading.Thread(target=self._server.serve_forever)
        self._t.setDaemon(True)  # don't hang on exit
        self._t.start()
        log.info("Listening on %s", self.address)
        if join:
            self._t.join()

    def stop(self):
        log.info("Closing server...")
        self._server.shutdown()
        self._server.server_close()
        # self._t.join()

    def update(self, data):
        self.layout.setBuffer(list(data))
        self.layout.push_to_driver()



"""
Convert back and forth between colors and string names.
"""

import numbers, re
from . import juce, classic


def name_to_color(name):
    """
    :param str name: a string identifying a color.  It might be a color name
                     from the two lists of color names, juce and classic; or
                     it might be a list of numeric r, g, b values separated by
                     commas.
    :returns: a color as an RGB 3-tuple
    """
    def to_color(name):
        name = name.lower()
        if ',' in name:
            if name.startswith('(') and name.endswith(')'):
                name = name[1:-1]

            r, g, b = name.split(',')
            return _from_number(r), _from_number(g), _from_number(b)

        try:
            n = _from_number(name)
        except:
            return TO_COLOR[name.replace(' ', '').lower()]

        return _to_triplet(n)

    try:
        color = to_color(name)
    except:
        raise ValueError('Unknown color name %s' % str(name))

    if not all(0 <= i <= 255 for i in color):
        raise ValueError('Component out of range: %s' % color)

    return color


def color_to_name(color, use_hex=False):
    """
    :param tuple color: an RGB 3-tuple of integer colors
    :returns: a string name for this color

    ``name_to_color(color_to_name(c)) == c`` is guaranteed to be true (but the
    reverse is not true, because name_to_color is a many-to-one function).
    """
    if isinstance(color, list):
        color = tuple(color)
    elif not isinstance(color, tuple):
        raise ValueError('Not a color')

    if use_hex:
        return '#%02x%02x%02x' % color

    try:
        return TO_NAME[color]
    except:
        return str(color)


def toggle(s):
    """
    Toggle back and forth between a name and a tuple representation.

    :param str s: a string which is either a text name, or a tuple-string:
                  a string with three numbers separated by commas

    :returns: if the string was a text name, return a tuple.  If it's a
              tuple-string and it corresponds to a text name, return the text
              name, else return the original tuple-string.
    """
    is_numeric = ',' in s or s.startswith('0x') or s.startswith('#')
    c = name_to_color(s)
    return color_to_name(c) if is_numeric else str(c)


class _Colors(object):
    def __getitem__(self, name):
        try:
            return name_to_color(name)
        except:
            raise KeyError(name)

    def __getattr__(self, name):
        try:
            return name_to_color(name)
        except:
            raise AttributeError("COLORS has no attribute' %s'" % name)

    def __call__(self, color):
        return color_to_name(color)


def _to_triplet(color):
    rg, b = color // 256, color % 256
    r, g = rg // 256, rg % 256
    return r, g, b


def _classic_pairs():
    find = re.compile(r'[A-Z][a-z0-9]+').finditer

    for name in dir(classic):
        if name == 'Green_HTML':
            yield 'green html', classic.Green_HTML

        elif len(name) > 1 and name[0].isupper() and name[1].islower():
            key = ' '.join(i.group(0).lower() for i in find(name))
            yield key, getattr(classic, name)


def _from_number(s):
    s = s.strip()
    for prefix in '0x', '#':
        if s.startswith(prefix):
            return int(s[len(prefix):], 16)

    return int(s)


def to_color(c):
    """Try to coerce the argument into a color - a 3-tuple of numbers-"""
    if isinstance(c, numbers.Number):
        return c, c, c

    if not c:
        raise ValueError('Cannot create color from empty "%s"' % c)

    if isinstance(c, str):
        return name_to_color(c)

    if isinstance(c, list):
        c = tuple(c)

    if isinstance(c, tuple):
        if len(c) > 3:
            return c[:3]
        while len(c) < 3:
            c += (c[-1],)
        return c

    raise ValueError('Cannot create color from "%s"' % c)


"""
COLOR is a "magic" color name variable.

 To get a color from a name, use ``COLORS.<colorname>`` - for example

::

    COLORS.red
    COLORS.ochre

or if the name is a variable or has a space in it,

::

    COLORS['violet red 4']

To get a name from a color, use

::

    COLOR((0, 255, 255))
"""
COLORS = _Colors()


_JUCE_COLORS = {k: _to_triplet(v) for k, v in juce.COLORS.items()}
_CLASSIC_COLORS = dict(_classic_pairs())

"""
A dictionary of every color by name.
"""

COLOR_DICT = dict(_CLASSIC_COLORS, **_JUCE_COLORS)

# TODO: integrate this better with COLOR
_SECONDARY_NAMES = juce.SECONDARY_NAMES.union({'off', 'on'})
TO_NAME = {v: k for k, v in COLOR_DICT.items() if k not in _SECONDARY_NAMES}
TO_COLOR = {k.replace(' ', ''): v for k, v in COLOR_DICT.items()}



"""
Test the all_pixel
"""

from .. drivers import ledtype
from .. util import log
from bibliopixel.animation.tests import StripChannelTest
from bibliopixel.layout.strip import Strip
from bibliopixel.drivers.serial import Serial
from bibliopixel.project.types import ledtype

DESCRIPTION = """
Equivalent to

.. code-block:: bash

    bp --num=10 --loglevel=debug --animation=strip_test --driver=serial \
      --fail_on_exception --layout=strip --ledtype=<argument>

"""

LEDTYPES = """
BiblioPixel currently understands the following types of LED strips:

""" + ', '.join(sorted(ledtype.LEDTYPE.__members__.keys()))

LEDTYPE_HELP = """\
The type of the LED strip that is connected to your AllPixel
""" + LEDTYPES

NO_LED_ERROR = """ERROR: No ledtype provided
""" + LEDTYPES


def run(args):
    if not args.ledtype:
        log.error(NO_LED_ERROR)
        return -1

    log.set_log_level('DEBUG')

    driver = Serial(ledtype=ledtype.make(args.ledtype), num=10)
    layout = Strip([driver])
    animation = StripChannelTest(layout)
    animation._set_runner(None)
    animation.start()


def add_arguments(args):
    args.set_defaults(run=run)
    args.add_argument('ledtype', help=LEDTYPE_HELP, nargs='?')



from .. channel_order import ChannelOrder
from . base import SPIBase
from ... util.colors import gamma

SPI_SPEED_ERROR = 'WS2801 requires an SPI speed of 1MHz but was set to {}MHz'


class WS2801(SPIBase):
    """Main driver for WS2801 based LED strips on devices like the
    Raspberry Pi and BeagleBone

    Provides the same parameters as
    :py:class:`bibliopixel.drivers.SPI.SPIBase`
    """

    def __init__(self, num, gamma=gamma.WS2801, spi_speed=1, **kwargs):
        if not (0 < spi_speed <= 1):
            raise ValueError(SPI_SPEED_ERROR.format(spi_speed))

        super().__init__(num, gamma=gamma, spi_speed=spi_speed, **kwargs)



import contextlib, threading, time
from . import adaptor, animation_threading, runner
from .. util import deprecated, log
from .. util.colors import palettes, legacy_palette
from .. project import attributes, fields


class Animation(object):
    free_run = False
    pre_recursion = fields.default_converter
    subframes = 1
    top_level = False

    FAIL_ON_EXCEPTION = False
    COLOR_DEFAULTS = ()

    @classmethod
    def construct(cls, project, *, run=None, name=None, data=None, **desc):
        """
        Construct an animation, set the runner, and add in the two
        "reserved fields" `name` and `data`.
        """
        from . failed import Failed
        exception = desc.pop('_exception', None)
        if exception:
            a = Failed(project.layout, desc, exception)
        else:
            try:
                a = cls(project.layout, **desc)
                a._set_runner(run or {})
            except Exception as e:
                if cls.FAIL_ON_EXCEPTION:
                    raise
                a = Failed(project.layout, desc, e)

        a.name = name
        a.data = data
        return a

    def __init__(self, layout, *, preclear=True, **kwds):
        self.palette = legacy_palette.pop_legacy_palette(
            kwds, *self.COLOR_DEFAULTS)
        self.palette.length = layout.numLEDs

        attributes.set_reserved(self, 'animation', **kwds)
        self.layout = layout
        assert layout
        self.internal_delay = None
        self.on_completion = None
        self.state = runner.STATE.ready
        self.preclear = preclear
        self.project = None
        self.runner = None
        self.time = time.time
        self.preframe_callbacks = []

    def set_project(self, project):
        self.project = project
        self.time = project.time
        self.runner.set_project(project)
        self.threading.set_project(project)

    @property
    def _led(self):
        """Many BiblioPixelAnimations use the "protected" variable _led."""
        return self.layout

    @_led.setter
    def _led(self, layout):
        self.layout = layout

    @property
    def color_list(self):
        return self.layout.color_list

    @color_list.setter
    def color_list(self, cl):
        self.layout.color_list = cl

    @property
    def completed(self):
        """Many BiblioPixelAnimations use the old `completed` variable."""
        return self.state == runner.STATE.complete

    @completed.setter
    def completed(self, state):
        self.state = runner.STATE.complete if state else runner.STATE.running

    def pre_run(self):
        pass

    @property
    def title(self):
        return self.__class__.__name__

    def step(self, amt=1):
        pass

    def cleanup(self, clean_layout=True):
        self.threading.cleanup()
        # Some cases we may not want to clear the screen
        # Like with the remote, it would flash between anims
        if clean_layout:
            self.layout.cleanup()

    def add_preframe_callback(self, callback):
        """
        The preframe_callbacks are called right before the start of a
        frame rendering pass.

        To avoid race conditions when editing values, the ``Project``
        adds a callback here for the top-level animation, to drain the
        edit_queue at a moment where no rendering is
        happening.
        """
        self.preframe_callbacks.append(callback)

    def start(self):
        self.threading.start()

    def stop(self):
        self.threading.stop_event.set()

    def join(self, timeout=None):
        self.threading.join(timeout)

    def run_all_frames(self):
        for i in self.generate_frames():
            pass

    def generate_frames(self, clean_layout=True):
        self._pre_run()
        try:
            if self.runner.repeats != 0:
                while self.state == runner.STATE.running:
                    self._run_one_frame()
                    yield
        finally:
            self.cleanup(clean_layout)

        self.on_completion and self.on_completion(self.state)
        self.state = runner.STATE.ready

    def _run_one_frame(self):
        if self.top_level:
            timestamps = [self.time()]
            for cb in self.preframe_callbacks:
                cb()

        self.step(self.runner.amt)

        if self.top_level:
            timestamps.append(self.time())

            self.layout.frame_render_time = timestamps[1] - timestamps[0]
            self.layout.push_to_driver()

            timestamps.append(self.time())
            _report_framerate(timestamps)

        self.cur_step += 1
        if self.state == runner.STATE.complete and self.runner.max_cycles > 0:
            if self.cycle_count < self.runner.max_cycles - 1:
                self.cycle_count += 1
                self.state = runner.STATE.running

        if self.top_level:
            self.threading.wait(self.sleep_time / self.subframes, timestamps)

        if self.threading.stop_event.isSet():
            self.state = runner.STATE.canceled
        else:
            self.state = self.runner.compute_state(self.cur_step, self.state)

    def _pre_run(self):
        self.state = runner.STATE.running
        self.runner.run_start_time = self.time()
        self.threading.stop_event.clear()

        self.cur_step = 0
        self.cycle_count = 0

        if self.free_run:
            self.sleep_time = 0
        elif self.internal_delay:
            self.sleep_time = self.internal_delay
        else:
            self.sleep_time = self.runner.sleep_time

        adaptor.adapt_animation_layout(self)
        self.preclear and self.layout.all_off()

        self.pre_run()

    def _set_runner(self, run):
        self.runner = runner.Runner(**(run or {}))
        self.threading = animation_threading.AnimationThreading(
            self.runner, self.run_all_frames)

    def run(self, **kwds):
        deprecated.deprecated('BaseAnimation.run')
        self._set_runner(kwds)
        self.layout.start()
        self.start()


if deprecated.allowed():  # pragma: no cover
    BaseAnimation = Animation


def _report_framerate(timestamps):
    total_time = timestamps[-1] - timestamps[0]
    fps = int(1.0 / max(total_time, 0.001))
    log.frame("%dms/%dfps / Frame: %dms / Update: %dms",
              1000 * total_time,
              fps,
              1000 * (timestamps[1] - timestamps[0]),
              1000 * (timestamps[2] - timestamps[1]))



import threading, uuid
from ... util import log
from . SimpleWebSocketServer import WebSocket, SimpleWebSocketServer


class Client(WebSocket):
    POSITION_START = bytearray([0x00, 0x00])
    PIXEL_START = bytearray([0x00, 0x01])

    def __init__(self, *args, server):
        super().__init__(*args)
        self.server = server
        self.connected = False
        log.debug('Server started...')

    def handleConnected(self):
        log.debug('Connected:{}'.format(self.address))
        self.connected = True
        self.server.add_client(self)

    def handleClose(self):
        self.server.remove_client(self)
        self.connected = False
        log.debug('Closed:{}'.format(self.address))

    def handleMessage(self):
        pass

    def update(self, pixels=None, positions=None):
        if self.connected:
            if pixels:
                self.sendFragmentStart(self.PIXEL_START)
                self.sendFragmentEnd(pixels)
            if positions:
                self.sendFragmentStart(self.POSITION_START)
                self.sendFragmentEnd(positions)


class Server:
    def __init__(self, port, selectInterval):
        self.clients = set()
        self.state = {}

        try:
            self.ws_server = SimpleWebSocketServer(
                '', port, Client, server=self, selectInterval=selectInterval)
        except PermissionError as e:
            if port < 1024:
                msg = PORT_BELOW_1024_ERROR.format(**locals())
                raise ValueError(msg, *e.args)
            raise e

        self.thread = threading.Thread(target=self.target, daemon=True)
        self.thread.start()

    def stop(self):
        self.ws_server.stop()

    def close(self):
        self.ws_server.close()

    def close(self):
        self.server.close()

    def is_alive(self):
        return self.thread.is_alive()

    def target(self):
        log.info('Starting WebSocket server thread...')
        try:
            self.ws_server.serveforever()
        except:
            pass
        log.info('WebSocket server closed')

    def update(self, **state):
        self.state.update(state)
        for client in self.clients.copy():
            client.update(**state)

    def add_client(self, client):
        self.clients.add(client)
        client.update(**self.state)

    def remove_client(self, client):
        try:
            self.clients.remove(client)
        except:
            pass


PORT_BELOW_1024_ERROR = """
SimPixel attempted to open a port at {port} and got a PermissionError.

Many systems disallow binding to any port below 1024 for non-root users.
Try again with a port number that's 1024 or greater.
"""



import abc


class Strip(abc.ABC):
    """Base class for contiguous strips.  You can also use a list as a Strip."""

    @abc.abstractmethod
    def __getitem__(self, index):
        """`index` must be an integer, not a slice."""
        pass

    @abc.abstractmethod
    def __setitem__(self, index, value):
        """`index` must be an integer, not a slice."""
        pass

    @abc.abstractmethod
    def __len__(self):
        pass


def fill(strip, item, start=0, stop=None, step=1):
    """Fill a portion of a strip from start to stop by step with a given item.
    If stop is not given, it defaults to the length of the strip.
    """
    if stop is None:
        stop = len(strip)

    for i in range(start, stop, step):
        strip[i] = item


def make_strip_coord_map(num, invert=False, offset=0):
    result = list(range(num))
    if offset:
        result = [i + offset for i in result]
    if invert:
        result = result[::-1]
    return result


def make_strip_coord_map_multi(config, invert=False):
    offset = 0
    result = []
    for c in config:
        result.extend(make_strip_coord_map(offset=offset, **c))
        offset = len(result)

    if invert:
        result = result[::-1]

    return result


def make_strip_coord_map_positions(num):
    return [[x, 0, 0] for x in range(num)]



"""
Print the process ID of the current bp instance that is running a project,
if any.

DEPRECATED: use ``$ bpa-pid``
"""

DESCRIPTION = """
Example:

.. code-block:: bash

    $ bp pid

"""

from .. util import log, pid_context


def run(args):
    try:
        log.printer(pid_context.get_pid(args.pid_filename))
    except:
        log.error('No bp process running')
        log.debug('Could not find file %s', args.pid_filename)
        return -1


def add_arguments(parser):
    pid_context.add_arguments(parser)
    parser.set_defaults(run=run)



import collections, threading
from . import runnable


class Task(object):
    def __init__(self, task=None, event=None):
        self.task = task or (lambda: None)
        self.event = event or threading.Event()

    def run(self, next_task):
        """Wait for the event, run the task, trigger the next task."""
        self.event.wait()
        self.task()
        self.event.clear()

        next_task.event.set()


class TaskThread(runnable.LoopThread):
    def __init__(self, producer_task, consumer_task, daemon=True, **kwds):
        super().__init__(daemon=daemon, **kwds)
        self.producer_task = producer_task
        self.consumer_task = consumer_task

    def produce(self):
        self.producer_task.run(self.consumer_task)

    def run_once(self):
        self.consumer_task.run(self.producer_task)



import math, threading, time
from . matrix import Matrix
from .. util import deprecated

if deprecated.allowed():  # pragma: no cover
    class POV(Matrix):
        def __init__(self, *args, **kwds):
            raise ValueError('layout.POV has been removed. Use animation.POV')

    LEDPOV = POV



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from setuptools.command.install_scripts import install_scripts
from setuptools.command.install import install as _install
import os, platform, sys

printer = print  # noqa: T001

SCRIPTS_TO_INSTALL = 'bp', 'bibliopixel', 'bp-color', 'bp-pid'


class InstallScripts(install_scripts):
    def run(self):
        install_scripts.run(self)
        if not os.name == 'nt':
            return

        for filepath in self.get_outputs():
            # If we can find an executable name in the #! top line of the script
            # file, make .bat wrapper for script.
            with open(filepath, 'rt') as fobj:
                first_line = fobj.readline().lower()

            if not (first_line.startswith('#!') and 'python' in first_line):
                printer(NO_PYTHON_ERROR % filepath)
                continue

            path, fname = os.path.split(filepath)
            froot, ext = os.path.splitext(fname)
            bat_file = os.path.join(path, froot + '.bat')
            bat_contents = BAT_TEMPLATE.replace('{FNAME}', fname)

            printer('Making %s wrapper for %s' % (bat_file, filepath))
            if self.dry_run:
                continue

            with open(bat_file, 'wt') as fobj:
                fobj.write(bat_contents)


# From here: http://pytest.org/2.2.4/goodpractises.html
class RunTests(TestCommand):
    DIRECTORY = 'test'

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [self.DIRECTORY]
        self.test_suite = True

    def run_tests(self):
        # Import here, because outside the eggs aren't loaded.
        import pytest
        errno = pytest.main(self.test_args)
        if errno:
            raise SystemExit(errno)


class RunBenchmark(RunTests):
    DIRECTORY = 'benchmark'


class RunCoverage(RunTests):
    def run_tests(self):
        import coverage
        cov = coverage.Coverage(config_file=True)

        cov.start()
        super().run_tests()
        cov.stop()

        cov.report(file=sys.stdout)
        coverage = cov.html_report(directory='htmlcov')
        fail_under = cov.get_option('report:fail_under')
        if coverage < fail_under:
            printer('ERROR: coverage %.2f%% was less than fail_under=%s%%' % (
                    coverage, fail_under))
            raise SystemExit(1)


def _get_version():
    from os.path import abspath, dirname, join
    filename = join(dirname(abspath(__file__)), 'bibliopixel', 'VERSION')
    return open(filename).read().strip()


INSTALLATION_ERROR = """INSTALLATION ERROR!

BiblioPixel v3 requires Python 3.4+ but
you are using version {0.major}.{0.minor}.{0.micro}

If you absolutely require using Python 2,
please install BiblioPixel v2.x using:

    > pip install "bibliopixel<3.0"

However we highly recommend using the latest BiblioPixel
(v3+) with Python 3.4+.
"""


if sys.version_info.major != 3:
    printer(INSTALLATION_ERROR.format(sys.version_info))
    sys.exit(1)

VERSION = _get_version()

with open('requirements.txt') as f:
    REQUIRED = f.read().splitlines()

setup(
    name='BiblioPixel',
    version=VERSION,
    description=(
        'BiblioPixel is a pure python library for manipulating a wide variety '
        'of LED strip based displays, both in strip and matrix form.'),
    long_description=open('README.rst').read(),
    author='Adam Haile',
    author_email='adam@maniacallabs.com',
    url='http://github.com/maniacallabs/bibliopixel/',
    license='MIT',
    packages=find_packages(exclude=['test']) + ['ui', 'scripts'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    tests_require=['pytest'],
    cmdclass={
        'benchmark': RunBenchmark,
        'coverage': RunCoverage,
        'test': RunTests,
        'install_scripts': InstallScripts
    },
    include_package_data=True,
    scripts=['scripts/' + s for s in SCRIPTS_TO_INSTALL],
    install_requires=REQUIRED
)

BAT_TEMPLATE = \
    r"""@echo off
REM wrapper to use shebang first line of {FNAME}

set mypath=%~dp0
set pyscript="%mypath%{FNAME}"
set /p line1=<%pyscript%

if "%line1:~0,2%" == "#!" (goto :goodstart)
echo First line of %pyscript% does not start with "#!"
exit /b 1

:goodstart
set py_exe=%line1:~2%
call "%py_exe%" %pyscript% %*
"""

NO_PYTHON_ERROR = """
WARNING: No #!python executable found in %s, skipping .bat wrapper'"""



from . animation import Animation
from . circle import Circle
from . game import BaseGameAnim
from . matrix import BaseMatrixAnim
from . cube import BaseCubeAnim
from . off import Off
from . sequence import Sequence
from . strip import BaseStripAnim
from . receiver import BaseReceiver
from . tests import StripChannelTest, MatrixChannelTest, MatrixCalibrationTest

from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    strip_test = StripChannelTest
    matrix_calibration = MatrixCalibrationTest
    matrix_test = MatrixChannelTest
    from . off import OffAnim
    from . animation import BaseAnimation
    from . circle import BaseCircleAnim



from . address import number, Address
from . receiver import Receiver
from .. util import deprecated


class Editor(Receiver):
    """
    A `Editor` is a `Receiver` which gets and sets `Address`es, perhaps using an
    an optional `EditQueue` for the setting.

    When the `set_project` method is called, `Editor` searches
    down through the address and stores the most recent `edit_queue`
    method it finds.
    """

    def __init__(self, address=None, project=None):
        self.address = Address(address)
        if project is not None:
            self.set_project(project)

    def set_project(self, project):
        self.project = project
        self.segments = self.address.segments[:]
        self.last_segment = self.address.segments.pop()
        edit_queue = None

        for segment in self.address.segments:
            edit_queue = getattr(project, 'edit_queue', edit_queue)
            project = segment.get(project)

        self.edit_queue = edit_queue

    def receive(self, msg):
        """
        Receives a message, and either sets it immediately, or puts it on the
        edit queue if there is one.

        """
        if self.edit_queue:
            self.edit_queue.put_edit(self._set, msg)
        else:
            self._set(msg)

    def get(self):
        return self.last_segment.get(self._get())

    def set(self, value):
        self.receive((number(value),))

    def __bool__(self):
        return bool(self.address)

    def __str__(self):
        return str(self.address)

    def _get(self):
        project = self.project
        for segment in self.address.segments:
            project = segment.get(project)
        return project

    def _set(self, values):
        args = self.address.assignment + values
        self.last_segment.set(self._get(), *args)

    if deprecated.allowed():
        set_root = set_project

        @property
        def root(self):
            return self.project



PERMISSION_ERROR = """Cannot access SPI device.
Please see

    https://github.com/maniacallabs/bibliopixel/wiki/SPI-Setup

for details.
"""

CANT_FIND_ERROR = """Cannot find SPI device.
Please see

    https://github.com/maniacallabs/bibliopixel/wiki/SPI-Setup

for details.
"""

CANT_IMPORT_PERIPHERY_ERROR = """
Unable to import periphery. Please install:
    pip install python-periphery

Please see
    https://github.com/maniacallabs/bibliopixel/wiki/SPI-Setup

for details.
"""

BAD_FORMAT_ERROR = """
When using py-spidev, `dev` must be in the format /dev/spidev*.*
Please see https://github.com/maniacallabs/bibliopixel/wiki/SPI-Setup
"""

CANT_IMPORT_SPIDEV_ERROR = """
Unable to import spidev. Please install:

    pip install spidev
"""



import serial, serial.tools.list_ports
from distutils.version import LooseVersion
from . codes import CMDTYPE, LEDTYPE, SPIChipsets, BufferChipsets
from . import io
from ... drivers.return_codes import raise_error
from ... util import log, util
from ... project.importer import import_module


class Devices(object):
    """Manage a list of serial devices.

    :param str hardware_id: A valid USB VID:PID pair such as "1D50:60AB"
    :param int baudrate: Baud rate to connect to serial device
    """

    def __init__(self, hardware_id, baudrate):
        self.hardware_id = hardware_id
        self.baudrate = baudrate

    def find_serial_devices(self):
        """Scan and report all compatible serial devices on system.

        :returns: List of discovered devices
        """
        self.devices = {}
        hardware_id = "(?i)" + self.hardware_id  # forces case insensitive

        for ports in serial.tools.list_ports.grep(hardware_id):
            port = ports[0]
            try:
                id = self.get_device_id(port, self.baudrate)
                ver = self._get_device_version(port, self.baudrate)
            except:
                log.debug('Error getting device_id for %s, %s',
                          port, self.baudrate)
                if True:
                    raise
                continue

            if getattr(ports, '__len__', lambda: 0)():
                log.debug('Multi-port device %s:%s:%s with %s ports found',
                          self.hardware_id, id, ver, len(ports))
            if id < 0:
                log.debug('Serial device %s:%s:%s with id %s < 0',
                          self.hardware_id, id, ver)
            else:
                self.devices[id] = port, ver

        return self.devices

    def get_device(self, id=None):
        """Returns details of either the first or specified device

        :param int id: Identifier of desired device. If not given, first device
            found will be returned

        :returns tuple: Device ID, Device Address, Firmware Version
        """
        if id is None:
            if not self.devices:
                raise ValueError('No default device for %s' % self.hardware_id)
            id, (device, version) = sorted(self.devices.items())[0]

        elif id in self.devices:
            device, version = self.devices[id]

        else:
            error = 'Unable to find device with ID %s' % id
            log.error(error)
            raise ValueError(error)

        log.info("Using COM Port: %s, Device ID: %s, Device Ver: %s",
                 device, id, version)
        return id, device, version

    def error(self, fail=True, action=''):
        """
        SHOULD BE PRIVATE METHOD
        """
        e = 'There was an unknown error communicating with the device.'
        if action:
            e = 'While %s: %s' % (action, e)
        log.error(e)
        if fail:
            raise IOError(e)

    def set_device_id(self, dev, id, baudrate=921600):
        """Set device ID to new value.

        :param str dev: Serial device address/path
        :param id: Device ID to set
        :param baudrate: Baudrate to use when connectinh (optional)
        """
        if id < 0 or id > 255:
            raise ValueError("ID must be an unsigned byte!")

        com, code, ok = io.send_packet(CMDTYPE.SETID, 1, dev, baudrate, 5, id)
        if not ok:
            raise_error(code)

    def get_device_id(self, dev, baudrate=921600):
        """Get device ID at given address/path.

        :param str dev: Serial device address/path
        :param baudrate: Baudrate to use when connecting (optional)
        """
        com, code, ok = io.send_packet(CMDTYPE.GETID, 0, dev, baudrate, 5)
        if code is None:
            self.error(action='get_device_id')
        return code

    def _get_device_version(self, dev, baudrate=921600):
        com, code, ok = io.send_packet(CMDTYPE.GETVER, 0, dev, baudrate, 0.5)
        return ok and io.read_byte(com) or 0



import socket, sys, time, os

from . driver_base import DriverBase
from .. util import log, util
from .. drivers.return_codes import RETURN_CODES


class CMDTYPE:
    SETUP_DATA = 1  # reserved for future use
    PIXEL_DATA = 2
    BRIGHTNESS = 3


class NetworkUDP(DriverBase):
    """Driver for communicating with another device on the network via UDP

    To be used with either
    BiblioPixelAnimations.receivers.GenericNetworkReceiver
    or :py:class:`.network_receiver.NetworkReceiver`

    Provides the same parameters of :py:class:`.driver_base.DriverBase` as
    well as those below:

    :param str host: Network hostname or IP address of receiver.
    :param int port: Network port of receiver
    :param bool broadcast: If True, broadcast to subnet instead of specific host
    """

    def __init__(self, num=0, width=0, height=0, host="localhost",
                 broadcast=False, port=3142, broadcast_interface='', **kwds):
        super().__init__(num, width, height, **kwds)

        self._host = host
        self._port = port
        self._sock = None
        self._broadcast = broadcast
        self._broadcast_interface = broadcast_interface

    def _generateHeader(self, cmd, size):
        packet = bytearray()
        packet.append(cmd)
        packet.append(size & 0xFF)
        packet.append(size >> 8)
        return packet

    def _connect(self):
        try:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            if self._broadcast:
                self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

            return self._sock
        except socket.gaierror:
            error = "Unable to connect to or resolve host: {}".format(
                self._host)
            log.error(error)
            raise IOError(error)

    def _compute_packet(self):
        self._render()
        count = self.bufByteCount()
        self._packet = util.generate_header(CMDTYPE.PIXEL_DATA, count)
        self._packet.extend(self._buf)

    # Push new data to strand
    def _send_packet(self):
        try:
            s = self._connect()
            s.sendto(self._packet, (self._host, self._port))

            s.close()

        except Exception as e:
            log.exception(e)
            error = "Problem communicating with network receiver!"
            log.error(error)
            raise IOError(error)



import numpy
from ctypes import c_float, c_uint8
from multiprocessing.sharedctypes import RawArray
from .. util import log
from . import importer


# These are just the ones we are sure to work with.  numpy defines a
# lot more, and we pass the names right through, if you want to experiment
# with 'float128' or whatever.
NUMPY_TYPES = (
    'int', 'int8', 'int16', 'int32', 'int64',
    'uint', 'uint8', 'uint16', 'uint32', 'uint64',
    'float', 'float32', 'float64')

NUMPY_DEFAULTS = (True, 'true', 'True', 'float')


class Maker:
    def __init__(self, floating=None, shared_memory=False, numpy_dtype=None):
        if numpy_dtype:
            log.debug('Using numpy')
            if numpy_dtype in NUMPY_DEFAULTS:
                numpy_dtype = 'float32'
            if numpy_dtype not in numpy.sctypeDict:
                raise ValueError(BAD_NUMPY_TYPE_ERROR % numpy_dtype)

        if shared_memory and numpy_dtype:
            log.error('Shared memory for numpy arrays is not yet supported.')
            numpy_dtype = None

        if floating is None:
            floating = not shared_memory

        c_type = c_float if floating else c_uint8

        if shared_memory:
            self.bytes = lambda size: RawArray(c_uint8, size)
            self.color_list = lambda size: RawArray(3 * c_type, size)
            # Note https://stackoverflow.com/questions/37705974/

        elif numpy_dtype:
            self.bytes = bytearray
            self.color_list = lambda size: numpy.zeros((size, 3), numpy_dtype)

        else:
            self.bytes = bytearray
            self.color_list = lambda size: [(0, 0, 0)] * size


MAKER = Maker()
NUMPY_MAKER = Maker(numpy_dtype='float')

ColorList = MAKER.color_list
NumpyColorList = NUMPY_MAKER.color_list

BAD_NUMPY_TYPE_ERROR = """
Bad numpy_type "%s"

Possible numpy_type values include:
    """ + ' '.join(NUMPY_TYPES)



from . matrix import BaseMatrixAnim
from .. util import AttributeDict


class BaseGameAnim(BaseMatrixAnim):

    def __init__(self, layout, inputDev):
        super().__init__(layout)
        self._input_dev = inputDev
        self._keys = None
        self._lastKeys = None
        self._speedStep = 0
        self._speeds = {}
        self._keyfuncs = {}

    def _exit(self, type, value, traceback):
        if hasattr(self._input_dev, 'setLightsOff'):
            self._input_dev.setLightsOff(5)
        self._input_dev.close()

    def setSpeed(self, name, speed):
        self._speeds[name] = speed

    def getSpeed(self, name):
        return self._speeds.get(name)

    def _checkSpeed(self, speed):
        return not (self._speedStep % speed)

    def checkSpeed(self, name):
        return name in self._speeds and self._checkSpeed(self._speeds[name])

    def addKeyFunc(self, key, func, speed=1, hold=True):
        if not isinstance(key, list):
            key = [key]
        for k in key:
            self._keyfuncs[k] = AttributeDict({
                "func": func,
                "speed": speed,
                "hold": hold,
                "last": False,
                "inter": False
            })

    def handleKeys(self):
        kf = self._keyfuncs
        for key in self._keys:
            val = self._keys[key]
            if key in kf:
                cfg = kf[key]
                speed_pass = self._checkSpeed(cfg.speed)

                if cfg.hold:
                    if speed_pass:
                        if (val or cfg.inter):
                            cfg.func()
                        else:
                            cfg.inter = cfg.last = val
                elif speed_pass:
                    if (val or cfg.inter) and not cfg.last:
                        cfg.func()
                    cfg.inter = cfg.last = val
                else:
                    cfg.inter |= val
        self._lastKeys = self._keys

    def step(self, amt):
        self._keys = self._input_dev.getKeys()
        self._speedStep += 1



import functools, numbers
from ... util import duration


USAGE = """
A time can be initialized with:

* A number that is greater than or equal to zero, representing a time in
  seconds.

* A string:

2 weeks, 1 day, 1 hours, 3 minutes, 45 seconds, 340.4ms..."""


@functools.singledispatch
def make(c):
    raise ValueError("Don't understand type %s" % type(c), USAGE)


@make.register(numbers.Number)
def _(t):
    if t < 0:
        raise ValueError('Times must not be negative', USAGE)
    return t


@make.register(str)
def _(s):
    return duration.parse(s)



from . wrapper import Wrapper
from .. layout import matrix

# Takes a matrix animation and displays it as individual columns or rows over
# time on a Strip


class POV(Wrapper):
    def __init__(self, *args, is_vertical=True, serpentine=False, **kwds):
        super().__init__(*args, **kwds)
        w, h = self.animation.width, self.animation.height
        color_list = self.layout.maker.color_list(w * h)
        self.animation.layout = matrix.Matrix(
            self.layout.drivers, w, h,
            serpentine=serpentine, color_list=color_list)
        self.get = self.animation.layout.get

        # Is the display strip oriented vertically?
        self.is_vertical = is_vertical

        # self.actual is the number of actual pixels in the strip
        # self.subframes is the number of virtual subframes
        actual, self.subframes = (w, h) if is_vertical else (h, w)

        if actual != self.layout.numLEDs:
            dim = ('vertical', 'horizontal')[is_vertical]
            raise ValueError('%s dimension is %d but should match shape=%d' %
                             (dim, actual, self.layout.numLEDs))

    def step(self, amt=1):
        substep = self.cur_step % self.subframes
        if not substep:
            super().step()

        for i in range(self.layout.numLEDs):
            x, y = (substep, i) if self.is_vertical else (i, substep)
            self.layout.set(i, self.get(x, y))



from ... util import int_names

USAGE = """
An int_name is either an integer, a string representing an integer, or a
fanciful name that's either a day of week, a month, a planet, or a chemical
element.
"""


def make(c):
    try:
        return int_names.to_index(c)
    except:
        raise ValueError('Don\'t understand "%s"\n%s' % (c, USAGE))



import argparse
from .. util import deprecated, log

ARGS = None


def set_args(description, argv, *argument_addders):
    parser = argparse.ArgumentParser(description=description)
    for adder in (deprecated, log) + argument_addders:
        adder.add_arguments(parser)

    global ARGS
    ARGS = parser.parse_args(argv)
    log.apply_args(ARGS)
    return ARGS



from .. import wrapper
from ... project import importer
from ... layout import cutter
from . import functions
from ... util import class_name


class MatrixIndexer:
    cutter_class = cutter.Indexer

    def __init__(self, layout, by_row=True, function=None):
        self.layout = layout
        self.cutter = self.cutter_class(layout, by_row)
        self.function = function or functions.sorter

    def __call__(self):
        self.cutter.apply(self.function)


class Reprocess(wrapper.Wrapper):
    def __init__(self, *args, process=None, **kwds):
        super().__init__(*args, **kwds)
        self.preclear = False

        pname = class_name.class_name(MatrixIndexer)

        process = process or pname
        if isinstance(process, str):
            typename = process
            process = {}
        else:
            typename = process.pop('typename', pname)

        datatype = importer.import_symbol(
            typename, 'bibliopixel.animation.reprocess')
        self.process = datatype(layout=self.layout, **process)

    def step(self, amt=1):
        super().step(amt)
        self.process()



import functools

NAMES = {
    'up': (0, -1),
    'down': (0, 1),
    'left': (-1, 0),
    'right': (1, 0),
}


@functools.singledispatch
def make(c):
    raise ValueError(ERROR % c)


@make.register(str)
def _(c):
    items = [v for k, v in NAMES.items() if k.startswith(c)]
    if len(items) != 1:
        raise ValueError(ERROR % c)
    return items[0]


@make.register(tuple)
def _(c):
    if c not in NAMES.values():
        raise ValueError(ERROR % (c,))
    return c


@make.register(list)
def _(c):
    return make(tuple(c))


USAGE = """Direction values can be
* a string 'up', 'down', 'left', 'right', or
* a string prefix like 'u' or 'ri'
* one of four numeric pairs (0, -1), (0, 1), (-1, 0), (1, 0).
"""

ERROR = 'Don\'t understand direction "%s"\n' + USAGE



"""
Send a restart signal to a BiblioPixel process running on this
machine.

DEPRECATED: use

.. code-block:: bash

    $ kill -hup `bpa-pid`


"""

DESCRIPTION = """
Example: ``$ bp restart``

"""

import os, signal
from .. project import defaults
from .. util import log, pid_context

CHOICES = sorted(d for d in dir(signal) if d.startswith('SIG') and '_' not in d)


def run(args):
    try:
        pid = pid_context.get_pid(args.pid_filename)
    except:
        log.error('No bp process running')
        log.debug('Could not find file %s', args.pid_filename)
        return -1

    try:
        os.kill(pid, getattr(signal, args.signal))
    except:
        log.error('Failed to send signal %s to bp process %s', args.signal, pid)
        return -1

    log.info('%s sent to bp process %s', args.signal, pid)


def add_signal_arguments(parser, default, help=''):
    parser.add_argument(
        'signal', nargs='?', default=default, choices=CHOICES,
        help='Signal to send.' + help)
    pid_context.add_arguments(parser)
    parser.set_defaults(run=run)


def add_arguments(parser):
    add_signal_arguments(parser, 'SIGHUP', ' Default SIGHUP restarts bp.')



from .. layout import Circle, Cube, Matrix, Strip
from .. util import colors, log

BLACK = colors.COLORS.Black

LAYOUT_WARNING = """\
Animation %s expects a layout of type %s but layout is of type %s:
you might get unpredictable results.
"""


def adapt_animation_layout(animation):
    """
    Adapt the setter in an animation's layout so that Strip animations can run
    on on Matrix, Cube, or Circle layout, and Matrix or Cube animations can run
    on a Strip layout.
    """
    layout = animation.layout
    required = getattr(animation, 'LAYOUT_CLASS', None)

    if not required or isinstance(layout, required):
        return

    msg = LAYOUT_WARNING % (
        type(animation).__name__, required.__name__, type(layout).__name__)

    setter = layout.set
    adaptor = None

    if required is Strip:
        if isinstance(layout, Matrix):
            width = layout.width

            def adaptor(pixel, color=None):
                y, x = divmod(pixel, width)
                setter(x, y, color or BLACK)

        elif isinstance(layout, Cube):
            lx, ly = layout.x, layout.y

            def adaptor(pixel, color=None):
                yz, x = divmod(pixel, lx)
                z, y = divmod(yz, ly)
                setter(x, y, z, color or BLACK)

        elif isinstance(layout, Circle):

            def adaptor(pixel, color=None):
                layout._set_base(pixel, color or BLACK)

    elif required is Matrix:
        if isinstance(layout, Strip):
            width = animation.width

            def adaptor(x, y, color=None):
                setter(x + y * width, color or BLACK)

    if not adaptor:
        raise ValueError(msg)

    log.warning(msg)
    animation.layout.set = adaptor



import copy
from . update_threading import UpdateThreading
from .. project import attributes, data_maker, fields
from .. util import colors, deprecated, log


class Layout(object):
    """
    Base Layer class. Use Strip, Matrix, Cube, or Circle instead!

    :param drivers: A list of drivers
    :param threadedUpdate: If True, updates to this layout are done in a
        separate thread
    :param brightness: An initial brightness value from 0 to 255
    :param maker: A data maker to make color_lists. TODO: Link to what a maker
        is.
    :param color_list: If non-Null, the layout uses this color_list instead
        of creating its own

    :ivar int numLEDs: Total number of pixels held by this layout instance
    """

    CLONE_ATTRS = 'maker', 'brightness'

    pre_recursion = fields.default_converter

    @classmethod
    def construct(cls, project, **desc):
        """Construct a layout.
        SHOULD BE PRIVATE
        """
        return cls(project.drivers, maker=project.maker, **desc)

    def __init__(self, drivers, threadedUpdate=False, brightness=255,
                 maker=data_maker.MAKER, color_list=None, **kwds):
        attributes.set_reserved(self, 'layout', **kwds)
        self.drivers = drivers if isinstance(drivers, list) else [drivers]
        self.maker = maker

        # self._colors will always be the same list - i.e. is guaranteed only
        # to be changed by list surgery, never assignment.
        if color_list is None:
            if not hasattr(self, 'numLEDs'):
                self.numLEDs = sum(d.numLEDs for d in self.drivers)
            self._colors = maker.color_list(self.numLEDs)
        else:
            self.numLEDs = len(color_list)
            self._colors = color_list

        pos = 0
        for d in self.drivers:
            d.set_colors(self._colors, pos)
            pos += d.numLEDs

        self.frame_render_time = 0

        self.threading = UpdateThreading(threadedUpdate, self)
        self.brightness = brightness
        self.set_brightness(brightness)

    def set_pixel_positions(self, pixel_positions):
        """SHOULD BE PRIVATE"""
        for d in self.drivers:
            d.set_pixel_positions(pixel_positions)

    def update(self):
        """DEPRECATED: Use :py:func:`push_to_driver` instead"""
        deprecated.deprecated('Layout.update')
        return self.push_to_driver()

    def start(self):
        for d in self.drivers:
            d.start()

    def stop(self):
        self.threading.stop()
        for d in self.drivers:
            d.stop()

    def join(self, timeout=None):
        self.threading.wait()
        for d in self.drivers:
            d.join(timeout)

    def cleanup_drivers(self):
        for d in self.drivers:
            d.cleanup()

    def cleanup(self):
        self.all_off()
        try:
            self.push_to_driver()
        except:
            pass
        self.threading.wait_for_update()

    def clone(self):
        """
        Return an independent copy of this layout with a completely separate
        color_list and no drivers.
        """
        args = {k: getattr(self, k) for k in self.CLONE_ATTRS}
        args['color_list'] = copy.copy(self.color_list)
        return self.__class__([], **args)

    @property
    def shape(self):
        """
        Return a tuple indicating the dimensions of the layout - (x,) for a
        strip, (x, y) for an array, (x, y, z) for a cube, and
        (ring_count, ring_steps) for a circle.
        """
        raise NotImplementedError

    @property
    def dimensions(self):
        deprecated.deprecated('Layout.dimensions')
        return self.shape

    @property
    def color_list(self):
        return self._colors

    @color_list.setter
    def color_list(self, color_list):
        self.set_color_list(color_list)

    def set_color_list(self, color_list, offset=0):
        """
        Set the internal colors starting at an optional offset.

        If `color_list` is a list or other 1-dimensional array, it is reshaped
        into an N x 3 list.

        If `color_list` too long it is truncated; if it is too short then only
        the initial colors are set.
        """
        if not len(color_list):
            return
        color_list = colors.make.colors(color_list)

        size = len(self._colors) - offset
        if len(color_list) > size:
            color_list = color_list[:size]
        self._colors[offset:offset + len(color_list)] = color_list

    def _get_base(self, pixel):
        if pixel >= 0 and pixel < self.numLEDs:
            return self._colors[pixel]

        return colors.COLORS.Black  # don't go out of bounds

    def _set_base(self, pixel, color):
        if pixel >= 0 and pixel < self.numLEDs:
            if isinstance(color, str):
                color = colors.COLORS[color]
            else:
                color = tuple(color)
            self.color_list[pixel] = color

    def get_pixel_positions(self):
        result = []
        for x in range(len(self.numLEDs)):
            result.append([x, 0, 0])
        return result

    def push_to_driver(self):
        """Push the current pixel state to the driver"""
        # This is overridden elsewhere.
        self.threading.push_to_driver()

    def set_brightness(self, brightness):
        self.brightness = brightness
        for d in self.drivers:
            d.set_brightness(brightness)

    # Set single pixel to RGB value
    def setRGB(self, pixel, r, g, b):
        """Set single pixel using individual RGB values instead of tuple"""
        self._set_base(pixel, (r, g, b))

    def setHSV(self, pixel, hsv):
        """Set single pixel to HSV tuple"""
        color = colors.hsv2rgb(hsv)
        self._set_base(pixel, color)

    # turns off the desired pixel
    def setOff(self, pixel):
        """Set single pixel off"""
        self._set_base(pixel, (0, 0, 0))

    def all_off(self):
        """Set all pixels off"""
        self._colors[:] = [(0, 0, 0)] * self.numLEDs

    # Fill the strand (or a subset) with a single color using a Color object
    def fill(self, color, start=0, end=-1):
        """Fill the entire strip with RGB color tuple"""
        start = max(start, 0)
        if end < 0 or end >= self.numLEDs:
            end = self.numLEDs - 1
        for led in range(start, end + 1):  # since 0-index include end in range
            self._set_base(led, color)

    # Fill the strand (or a subset) with a single color using RGB values
    def fillRGB(self, r, g, b, start=0, end=-1):
        """Fill entire strip by giving individual RGB values instead of tuple"""
        self.fill((r, g, b), start, end)

    # Fill the strand (or a subset) with a single color using HSV values
    def fillHSV(self, hsv, start=0, end=-1):
        """Fill the entire strip with HSV color tuple"""
        self.fill(colors.hsv2rgb(hsv), start, end)


class MultiLayout(Layout):
    CLONE_ATTRS = Layout.CLONE_ATTRS + ('gen_coord_map', 'coord_map')

    def __init__(self, *args, gen_coord_map=None, coord_map=None, **kwds):
        super().__init__(*args, **kwds)
        self.gen_coord_map = gen_coord_map
        if gen_coord_map:
            if coord_map:
                log.warning('Cannot set both coord_map and gen_coord_map')
            elif isinstance(gen_coord_map, dict):
                coord_map = self.gen_multi(**gen_coord_map)
            else:
                coord_map = self.gen_multi(gen_coord_map)
        self.coord_map = coord_map

    def gen_multi(self, *args, **kwds):
        raise NotImplementedError

    def set_colors(self, buf):
        """
        DEPRECATED: use self.color_list

        Use with extreme caution!
        Directly sets the internal buffer and bypasses all brightness and
        rotation control buf must also be in the exact format required by the
        display type.
        """
        deprecated.deprecated('layout.set_colors')
        if len(self._colors) != len(buf):
            raise IOError("Data buffer size incorrect! "
                          "Expected: {} bytes / Received: {} bytes"
                          .format(len(self._colors), len(buf)))
        self._colors[:] = buf

    def setBuffer(self, buf):
        deprecated.deprecated('layout.setBuffer')
        self.color_list = buf



import contextlib, os, tempfile
from . import log

DEFAULT_PID_FILENAME = os.path.join(tempfile.gettempdir(), 'bp_pid_file.txt')
HELP = 'Filename to store the `bp` process ID when running'


def add_arguments(parser):
    parser.add_argument(
        '--pid_filename', default=DEFAULT_PID_FILENAME, help=HELP)


@contextlib.contextmanager
def pid_context(pid_filename=DEFAULT_PID_FILENAME):
    """
    For the duration of this context manager, put the PID for this process into
    `pid_filename`, and then remove the file at the end.
    """
    if os.path.exists(pid_filename):
        contents = open(pid_filename).read(16)
        log.warning('pid_filename %s already exists with contents %s...',
                    pid_filename, contents)

    with open(pid_filename, 'w') as fp:
        fp.write(str(os.getpid()))
        fp.write('\n')

    try:
        yield
    finally:
        try:
            os.remove(pid_filename)
        except Exception as e:
            log.error('Got an exception %s deleting the pid_filename %s',
                      e, pid_filename)


def get_pid(pid_filename=DEFAULT_PID_FILENAME):
    """
    Return the integer PID for the current bp process, or raise an exception if
    there is no such process or it hasn't registered a PID.
    """
    return int(open(pid_filename).read(16))



import threading, time, webbrowser
from ... util import log

WEBPAGE_OPENED = False


def opener(ip_address, port, delay=1):
    """
    Wait a little and then open a web browser page for the control panel.
    """
    global WEBPAGE_OPENED
    if WEBPAGE_OPENED:
        return
    WEBPAGE_OPENED = True
    raw_opener(ip_address, port, delay)


def raw_opener(ip_address, port, delay=1):
    """
    Wait a little and then open a web browser page for the control panel.
    """
    def target():
        time.sleep(delay)
        url = 'http://%s:%d' % (ip_address, port)
        webbrowser.open(url, new=0, autoraise=True)

    threading.Thread(target=target, daemon=True).start()



def color_blend(a, b):
    """
    Performs a Screen blend on RGB color tuples, a and b
    """
    return (255 - (((255 - a[0]) * (255 - b[0])) >> 8),
            255 - (((255 - a[1]) * (255 - b[1])) >> 8),
            255 - (((255 - a[2]) * (255 - b[2])) >> 8))


def color_scale(color, level):
    """
    Scale RGB tuple by level, 0 - 256
    """
    return tuple([int(i * level) >> 8 for i in list(color)])



"""
More or less uniformly run something as a new daemon thread or process.
"""

import multiprocessing, threading, queue


def _run_locally(input, output, function, args, **kwds):
    function(input, output, *args, **kwds)


def run(function, *args, use_subprocess=False, daemon=True, **kwds):
    """
    Create input, output queues, call `function` in a subprocess or a thread.

    ``function`` is called like this: ``function(input, output, *args, **kwds)``

    :param use_subprocess: if true, create a new multiprocess;
                           if false, create a new thread
    :param function: the function to call
    :param daemon: is the thread or subprocess run as a daemon or not?

    :param args: positional arguments to the function
    :param kwds: keyword arguments to the function
    :returns: a tuple with three elements: the subprocess or thread, an input
              queue, and an output queue.
    """
    if use_subprocess:
        Creator, Queue = multiprocessing.Process, multiprocessing.Queue
    else:
        Creator, Queue = threading.Thread, queue.Queue

    input, output = Queue(), Queue()
    args = input, output, function, args
    sub = Creator(target=_run_locally, args=args, kwargs=kwds, daemon=daemon)
    sub.start()

    return sub, input, output
